@@ -2,7 +2,11 @@ from .agent_base import AgentBase
 # from .random_walk import RandomWalker
 import random
 import pandas as pd
+from fractions import Fraction
+from decimal import *
 
+def Solidity(val):
+    return int(Decimal(val) * 10 ** 18)
 
 class LiquidityProviderAgent(AgentBase):
     """ subclass of RandomWalker, which is subclass to Mesa Agent
@@ -30,16 +34,22 @@ class LiquidityProviderAgent(AgentBase):
 
     def deposit(self, tkn, amount, block_num=0, fake_wallet=True):
 
+        amount = Solidity(amount)
+
         if fake_wallet:
             self.handle_fake_wallet(tkn, amount, m=1000)
         else:
-            self.wallet[f'bn{tkn}'] = [0]
+            if (f'bn{tkn}' not in self.wallet):
+                self.wallet[f'bn{tkn}'] = [Solidity(0)]
+            if (tkn == 'BNT') & (f'v{tkn}' not in self.wallet):
+                self.wallet[f'v{tkn}'] = [Solidity(0)]
 
         for key in self.wallet:
             if (tkn == key) & (tkn == 'BNT'):
                 self.wallet[tkn].append(self.wallet[tkn][-1] - amount)
-                self.wallet[f"v{tkn}"].append(self.wallet[f"v{tkn}"][-1] + self.protocol.stake(tkn, amount, block_num))
-                self.wallet[f"bn{tkn}"].append(self.wallet[f"bn{tkn}"][-1] + self.protocol.stake(tkn, amount, block_num))
+                amt = self.protocol.stake(tkn, amount, block_num)
+                self.wallet[f"v{tkn}"].append(self.wallet[f"v{tkn}"][-1] + amt)
+                self.wallet[f"bn{tkn}"].append(self.wallet[f"bn{tkn}"][-1] + amt)
             elif tkn == key:
                 self.wallet[tkn].append(self.wallet[tkn][-1] - amount)
                 self.wallet[f"bn{tkn}"].append(self.wallet[f"bn{tkn}"][-1] + self.protocol.stake(tkn, amount, block_num))
@@ -56,7 +66,7 @@ class LiquidityProviderAgent(AgentBase):
 
         if len(self.wallet['BNT']) < len(self.wallet[tkn]):
             self.wallet['BNT'].append(self.wallet['BNT'][-1])
-        
+
         # if len(self.wallet['bnBNT']) < len(self.wallet[tkn]):
         #     self.wallet['bnBNT'].append(self.wallet['bnBNT'][-1])
         # if len(self.wallet['bnTKN']) < len(self.wallet[tkn]):
