@@ -1,6 +1,6 @@
 from bancorml.environments.env_base import EnvBase
 from bancorml.agents import LiquidityProviderAgent
-
+from fractions import Fraction
 
 class Bancor3(EnvBase):
     """Bancor3 Environment.
@@ -13,7 +13,7 @@ class Bancor3(EnvBase):
             pool_fees (dict): Dictionary of tkn:fee (str:float) key-values per pool.
             cooldown_period (int): The cooldown period in days.
             exit_fee (float): The global exit fee.
-            spot_prices (dict): Dictionary of tkn:price (str:float) key-values per tkn.
+            spot_rates (dict): Dictionary of tkn:price (str:float) key-values per tkn.
             whitelisted_tokens (list): List of token tickernames indicating whitelist status approval.
             bootstrapped_tokens (list): List of DAOmsig initiated pools/tokens.
             vortex_rates (dict): Dictionary of tkn:rate (str:float) key-values per TKN.
@@ -33,58 +33,58 @@ class Bancor3(EnvBase):
     def __init__(
             self,
             block_num=0,
-            alpha=0.2,
+            alpha=Fraction(2, 10),
             is_solidity=False,
             min_liquidity_threshold=10000,
-            exit_fee=.0025,
+            exit_fee=Fraction(2500, 1000000),
             cooldown_period=7,
             bnt_funding_limit=100000,
-            external_price_feed={
+            external_rate_feed={
                         "TKN": {'block_num': [0],
-                                'price_usd': [7.50]},
+                                'rate': [Fraction(2.50 / 3.0)]},
                         "BNT": {'block_num': [0],
-                                'price_usd': [2.50]},
+                                'rate': [2.50]},
                         "LINK": {'block_num': [0],
-                                'price_usd': [15.00]},
+                                'rate': [15.00]},
                         "ETH": {'block_num': [0],
-                                'price_usd': [2500.00]},
+                                'rate': [2500.00]},
                         "wBTC": {'block_num': [0],
-                                'price_usd': [40000.00]},
+                                'rate': [40000.00]},
                         "DAI": {'block_num': [0],
-                                'price_usd': [1.00]}
+                                'rate': [1.00]}
             },
             pool_fees={
-                "TKN": 0.01,
-                "BNT": 0.01,
-                "LINK": 0.01,
-                "ETH": 0.01,
-                "wBTC": 0.01,
-                "DAI": 0.01,
+                "TKN": Fraction(1, 100),
+                "BNT": Fraction(1, 100),
+                "LINK": Fraction(1, 100),
+                "ETH": Fraction(1, 100),
+                "wBTC": Fraction(1, 100),
+                "DAI": Fraction(1, 100),
             },
             whitelisted_tokens=['wBTC','BNT','LINK','ETH','DAI','TKN'],
             bootstrapped_tokens=[],
             exchange_rates={},
-            spot_prices={
+            spot_rates={
                 "TKN": {'block_num': [0],
-                        'price_usd': [7.50]},
+                        'rate': [Fraction(2.50 / 3.0)]},
                 "BNT": {'block_num':[0],
-                        'price_usd':[2.50]},
+                        'rate':[2.50]},
                 "LINK": {'block_num':[0],
-                        'price_usd':[15.00]},
+                        'rate':[15.00]},
                 "ETH": {'block_num':[0],
-                        'price_usd':[2500.00]},
+                        'rate':[2500.00]},
                 "wBTC": {'block_num':[0],
-                        'price_usd':[40000.00]},
+                        'rate':[40000.00]},
                 "DAI": {'block_num': [0],
-                         'price_usd': [1.00]}
+                         'rate': [1.00]}
             },
             vortex_rates={
-                "TKN": 0.2,
-                "BNT": 0.2,
-                "LINK": 0.2,
-                "ETH": 0.2,
-                "wBTC": 0.2,
-                "DAI": 0.2,
+                "TKN": Fraction(2, 10),
+                "BNT": Fraction(2, 10),
+                "LINK": Fraction(2, 10),
+                "ETH": Fraction(2, 10),
+                "wBTC": Fraction(2, 10),
+                "DAI": Fraction(2, 10),
             },
             warning_only=False,
             dao_msig_initialized_pools=[],
@@ -96,7 +96,7 @@ class Bancor3(EnvBase):
         self._ema = Bancor3.ema
         self.set_param(alpha=alpha,
                        warning_only=warning_only,
-                       spot_prices=spot_prices,
+                       spot_rates=spot_rates,
                        vortex_rates=vortex_rates,
                        block_num=block_num,
                        exchange_rates=exchange_rates,
@@ -107,7 +107,7 @@ class Bancor3(EnvBase):
                        pool_fees=pool_fees,
                        bootstrapped_tokens=bootstrapped_tokens,
                        whitelisted_tokens=whitelisted_tokens,
-                       external_price_feed=external_price_feed,
+                       external_rate_feed=external_rate_feed,
                        is_solidity=is_solidity,
                        dao_msig_initialized_pools=dao_msig_initialized_pools,
                        verbose=verbose)
@@ -230,6 +230,10 @@ class Bancor3(EnvBase):
             bnbnt_supply = self.pool_token_supply_ledger['bnBNT_ERC20_contract']['supply'][-1]
             staked_bnt = self.staking_ledger['BNT'][-1]
             pool_tokens_issued_to_user = self.tokenomics.get_pool_tokens_issued_to_user(x, bnbnt_supply, staked_bnt)
+            update_amt = bnbnt_supply - pool_tokens_issued_to_user
+            self.pool_token_supply_ledger["bnBNT_ERC20_contract"]['block_num'].append(self.block_num)
+            self.pool_token_supply_ledger["bnBNT_ERC20_contract"]['supply'].append(update_amt)
+
         else:
             pool_tokens_issued_to_user = self.mint_bntkn(tkn, x)
             self.add_tkn_to_vault(tkn, x)
@@ -272,9 +276,19 @@ class Bancor3(EnvBase):
                 operation = self.tokenomics.handle_trade_tkn_to_bnt
 
             a = self.available_liquidity_ledger[tkn_a]['BNT'][-1]
+            a = self.validate_type(a)
             b = self.available_liquidity_ledger[tkn_a][tkn_a][-1]
+            b = self.validate_type(b)
             d = self.pool_fees[tkn_a]
+            d = self.validate_type(d)
             e = self.vortex_rates[tkn_a]
+            e = self.validate_type(e)
+
+            print(f'a={a}, \n'
+                  f'b={b}, \n'
+                  f'd={d}, \n'
+                  f'e={e}, \n'
+                  f'x={x}, \n')
             bnt_trading_liquidity, tkn_trading_liquidity, tkn_out, tkn_fee, vortex_fee = operation(
                 a, d, b, x, e)
 
@@ -282,9 +296,10 @@ class Bancor3(EnvBase):
                 self.update_available_liquidity_ledger(tkn_a, bnt_trading_liquidity, tkn_trading_liquidity)
 
             self.add_tkn_to_vault(tkn1, x)
-            self.add_tkn_to_vault(tkn2, -tkn_out)
+            self.remove_tkn_from_vault(tkn2, tkn_out)
             self.add_tkn_to_staking_ledger(tkn2, tkn_fee)
-            self.add_tkn_to_vortex_ledger(vortex_fee)
+            self.add_tkn_to_vortex_ledger(tkn2, vortex_fee)
+
 
             print(f'bnt_trading_liquidity={bnt_trading_liquidity}, \n'
                   f'tkn_trading_liquidity={tkn_trading_liquidity}, \n'
@@ -295,14 +310,21 @@ class Bancor3(EnvBase):
         else:
 
             a1 = self.available_liquidity_ledger[tkn1]['BNT'][-1]
+            a1 = self.validate_type(a1)
             b1 = self.available_liquidity_ledger[tkn1][tkn1][-1]
+            b1 = self.validate_type(b1)
             d1 = self.pool_fees[tkn1]
+            d1 = self.validate_type(d1)
 
             a2 = self.available_liquidity_ledger[tkn2]['BNT'][-1]
+            a2 = self.validate_type(a2)
             b2 = self.available_liquidity_ledger[tkn2][tkn2][-1]
+            b2 = self.validate_type(b2)
             d2 = self.pool_fees[tkn2]
+            d2 = self.validate_type(d2)
 
             e = self.vortex_rates[tkn1]
+            e = self.validate_type(e)
 
             bnt_source_trading_liquidity, tkn_source_trading_liquidity, bnt_destination_trading_liquidity, tkn_destination_trading_liquidity, tkn_out, bnt_fee, tkn_fee, vortex_fee = self.tokenomics.handle_trade_tkn_to_tkn(
                 a1, d1, b1, a2, d2, b2, x, e)
@@ -330,6 +352,6 @@ class Bancor3(EnvBase):
 
         self.check_pool_shutdown(warning_only=self.warning_only)
         for token in [tkn1, tkn2]:
-            self.update_spot_prices(token)
+            self.update_spot_rates(token)
         return tkn_out
 
