@@ -12,6 +12,7 @@ from bancorml.utils import FixedPointUnstakeTKN, FloatingPointUnstakeTKN, FixedP
 from bancorml.environments.tokenomics.floating_point_tokenomics import FloatingPointTokenomics
 from bancorml.environments.tokenomics.fixed_point_tokenomics import FixedPointTokenomics
 from fxpmath import Fxp
+from fractions import Fraction
 
 # logger = logging.getLogger(__name__)
 from bancorml.utils.schemas import FixedPoint
@@ -32,7 +33,7 @@ class EnvBase(ABC):
         # observation_space=observation_space,
     ):
         # self.observation_space = observation_space
-        # self._external_price_feed = external_price_feed
+        # self._external_rate_feed = external_rate_feed
         # self.logger = get_logger(f"{__name__}")
 
         # create a vault
@@ -54,12 +55,12 @@ class EnvBase(ABC):
         self.actions = 0
 
     @property
-    def min_liquidity_threshold(self) -> int or float:
+    def min_liquidity_threshold(self) -> int or float or Decimal:
         """Minimum pool liquidity threshold before the DAOmsig can initiate."""
         return self._min_liquidity_threshold
 
     @min_liquidity_threshold.setter
-    def min_liquidity_threshold(self, updated_min_liquidity_threshold: int or float):
+    def min_liquidity_threshold(self, updated_min_liquidity_threshold: int or float or Decimal):
         self._min_liquidity_threshold = updated_min_liquidity_threshold
 
     @property
@@ -72,21 +73,21 @@ class EnvBase(ABC):
         self._warning_only = updated_warning_only
 
     @property
-    def bnt_funding_limit(self) -> int or float:
+    def bnt_funding_limit(self) -> int or float or Decimal:
         """The BancorDAO determines the available liquidity for trading, through adjustment of the “BNT funding limit” parameter."""
         return self._bnt_funding_limit
 
     @bnt_funding_limit.setter
-    def bnt_funding_limit(self, updated_bnt_funding_limit: int or float):
+    def bnt_funding_limit(self, updated_bnt_funding_limit: int or float or Decimal):
         self._bnt_funding_limit = updated_bnt_funding_limit
 
     @property
-    def alpha(self) -> float:
+    def alpha(self) -> int or float or Decimal:
         """EMA calculation alpha setting"""
         return self._alpha
 
     @alpha.setter
-    def alpha(self, updated_alpha: float):
+    def alpha(self, updated_alpha: float or Decimal):
         self._alpha = updated_alpha
 
     @property
@@ -117,13 +118,13 @@ class EnvBase(ABC):
         self._dao_msig_initialized_pools = updated_dao_msig_initialized_pools
 
     @property
-    def external_price_feed(self) -> dict:
+    def external_rate_feed(self) -> dict:
         """List of approved tokens allowed on Bancor v3"""
-        return self._external_price_feed
+        return self._external_rate_feed
 
-    @external_price_feed.setter
-    def external_price_feed(self, updated_external_price_feed: dict):
-        self._external_price_feed = updated_external_price_feed
+    @external_rate_feed.setter
+    def external_rate_feed(self, updated_external_rate_feed: dict):
+        self._external_rate_feed = updated_external_rate_feed
 
     @property
     def bootstrapped_tokens(self) -> list:
@@ -153,7 +154,7 @@ class EnvBase(ABC):
         self._pool_fees = updated_pool_fees
 
     @property
-    def exit_fee(self) -> Decimal:
+    def exit_fee(self) -> int or float or Decimal:
         """
         As a final circumspection against bad behavior,
         exit fees are introduced into the Bancor ecosystem for the first time.
@@ -162,7 +163,7 @@ class EnvBase(ABC):
         return self._exit_fee
 
     @exit_fee.setter
-    def exit_fee(self, updated_exit_fee: Decimal):
+    def exit_fee(self, updated_exit_fee: float or Decimal):
         self._exit_fee = updated_exit_fee
 
     @property
@@ -184,13 +185,13 @@ class EnvBase(ABC):
         self._ema = updated_ema
 
     @property
-    def spot_prices(self) -> dict:
+    def spot_rates(self) -> dict:
         """Current system state spot prices. Setting these will cause an auto-update"""
-        return self._spot_prices
+        return self._spot_rates
 
-    @spot_prices.setter
-    def spot_prices(self, updated_spot_prices: dict):
-        self._spot_prices = updated_spot_prices
+    @spot_rates.setter
+    def spot_rates(self, updated_spot_rates: dict):
+        self._spot_rates = updated_spot_rates
 
     @property
     def custom_name(self):
@@ -210,11 +211,11 @@ class EnvBase(ABC):
                         pool_fees=None,
                         cooldown_period=None,
                         exit_fee=None,
-                        spot_prices=None,
+                        spot_rates=None,
                         whitelisted_tokens=None,
                         bootstrapped_tokens=None,
                         exchange_rates=None,
-                        external_price_feed=None,
+                        external_rate_feed=None,
                         vortex_rates=None,
                         dao_msig_initialized_pools=None,
                         warning_only=None,
@@ -230,7 +231,7 @@ class EnvBase(ABC):
             pool_fees (dict): Dictionary of tkn:fee (str:float) key-values per pool.
             cooldown_period (int): The cooldown period in days.
             exit_fee (float): The global exit fee.
-            spot_prices (dict): Dictionary of tkn:price (str:float) key-values per tkn.
+            spot_rates (dict): Dictionary of tkn:price (str:float) key-values per tkn.
             whitelisted_tokens (list): List of token tickernames indicating whitelist status approval.
             bootstrapped_tokens (list): List of DAOmsig initiated pools/tokens.
             vortex_rates (dict): Dictionary of tkn:rate (str:float) key-values per TKN.
@@ -265,24 +266,24 @@ class EnvBase(ABC):
             self.set_whitelisted_tokens(whitelisted_tokens)
         if vortex_rates is not None:
             self.vortex_rates = vortex_rates
-        if external_price_feed is not None:
-            self.external_price_feed = external_price_feed
+        if external_rate_feed is not None:
+            self.external_rate_feed = external_rate_feed
         if bootstrapped_tokens is not None:
             self.bootstrapped_tokens = bootstrapped_tokens
         if dao_msig_initialized_pools is not None:
             self.dao_msig_initialized_pools = dao_msig_initialized_pools
-        if spot_prices is not None:
-            self.spot_prices = spot_prices
+        if spot_rates is not None:
+            self.spot_rates = spot_rates
 
         self.update_exchange_rates()
         self.init_ema()
 
     def init_tokens(self, add_tkns=[]):
         whitelisted_tokens = self.whitelisted_tokens
-        external_price_feed = self.external_price_feed
+        external_rate_feed = self.external_rate_feed
         pool_fees = self.pool_fees
         vortex_rates = self.vortex_rates
-        spot_prices = self.spot_prices
+        spot_rates = self.spot_rates
         ema = self.ema
 
         for tkn in add_tkns:
@@ -290,30 +291,30 @@ class EnvBase(ABC):
                 whitelisted_tokens.append(tkn)
 
         for tkn in whitelisted_tokens:
-            if tkn not in external_price_feed:
+            if tkn not in external_rate_feed:
                 if self.verbose:
-                    print(f'WARNING: {tkn} was added to external_price_feed at 1.00 USD \n'
+                    print(f'WARNING: {tkn} was added to external_rate_feed at 1.00 USD \n'
                           f'You can update this with the .update_external_price(tkn, external_price) method. \n')
-                external_price_feed[tkn] =  {'block_num': [self.block_num],
-                                             'price_usd': [1.00]}
+                external_rate_feed[tkn] =  {'block_num': [self.block_num],
+                                             'rate': [1]}
             if tkn not in pool_fees:
                 if self.verbose:
                     print(f'WARNING: {tkn} was added to pool_fees at 0.01 \n'
                           f'You can update this with the .update_pool_fee(tkn, pool_fee) method. \n')
-                pool_fees[tkn] = 0.01
+                pool_fees[tkn] = Fraction(1, 100)
 
             if tkn not in vortex_rates:
                 if self.verbose:
                     print(f'WARNING: {tkn} was added to vortex_rates at 0.2 \n'
                           f'You can update this with the .update_vortex_rate(tkn, vortex_rate) method. \n')
-                vortex_rates[tkn] = 0.2
+                vortex_rates[tkn] = Fraction(2, 10)
 
-            if tkn not in spot_prices:
+            if tkn not in spot_rates:
                 if self.verbose:
-                    print(f'WARNING: {tkn} was added to spot_prices at $1 USD \n'
-                          f'You can update this with the .update_spot_price(tkn, price_usd) method. \n')
-                spot_prices[tkn] = {'block_num': [self.block_num],
-                                    'price_usd': [1.00]}
+                    print(f'WARNING: {tkn} was added to spot_rates at $1 USD \n'
+                          f'You can update this with the .update_spot_price(tkn, rate) method. \n')
+                spot_rates[tkn] = {'block_num': [self.block_num],
+                                    'rate': [1]}
 
             if tkn not in self.vault_ledger:
                 self.vault_ledger[tkn] = [0]
@@ -343,15 +344,15 @@ class EnvBase(ABC):
 
         self.ema = ema
         self.whitelisted_tokens = whitelisted_tokens
-        self.spot_prices = spot_prices
-        self.external_price_feed = external_price_feed
+        self.spot_rates = spot_rates
+        self.external_rate_feed = external_rate_feed
         self.vortex_rates = vortex_rates
         self.pool_fees = pool_fees
 
 
     def init_ema(self):
         ema = self.ema
-        for tkn in self.spot_prices:
+        for tkn in self.spot_rates:
             if tkn in ema:
                 ema[tkn]['block_num'] = [self.block_num]
                 ema[tkn]['ema'] = [self.exchange_rates[tkn]]
@@ -401,7 +402,7 @@ class EnvBase(ABC):
         """
         return pd.DataFrame.from_dict(self.pool_token_supply_ledger[f'bn{tkn}_ERC20_contract'])
 
-    def update_spot_prices(self, tkn):
+    def update_spot_rates(self, tkn):
         """Updates the internal spot price as determined by the trading liquidity balances of the pool
 
         Args:
@@ -409,12 +410,11 @@ class EnvBase(ABC):
         Returns:
             self
         """
-        spot_rate = self.get_spot_rate(tkn)
-        spot_prices = self.spot_prices
-        spot_prices[tkn]['block_num'].append(self.block_num)
-        spot_prices[tkn]['price_usd'].append(spot_rate)
-        self.spot_prices = spot_prices
-        # self.update_exchange_rates()
+        spot_rate = self.get_internal_spot_rate(tkn)
+        spot_rates = self.spot_rates
+        spot_rates[tkn]['block_num'].append(self.block_num)
+        spot_rates[tkn]['rate'].append(spot_rate)
+        self.spot_rates = spot_rates
 
     def set_whitelisted_tokens(self, whitelisted_tokens, bootstrap_tokens=True):
         """Check if the spot rate is outside of the EMA tolerance. No changes are made to the pool depth if outside limits.
@@ -423,7 +423,7 @@ class EnvBase(ABC):
             whitelisted_tokens (list): The tickername of the token.
             bootstrap_tokens (bool): If true, automatically initializes all whitelisted tokens and adds them to the bootstrapped_tokens list.
         Returns:
-            bnTKN_value (int or float or decimal): value of a single bnTKN pool token.
+            bnTKN_value (int or float or Decimal or decimal): value of a single bnTKN pool token.
         """
         self.whitelisted_tokens = whitelisted_tokens
         if bootstrap_tokens:
@@ -443,39 +443,60 @@ class EnvBase(ABC):
             self.staking_ledger[tkn] = [0]
             self.external_protection_wallet_ledger[tkn] = [0]
 
-    def validate_type(self, val):
+    def handle_split_decimal(self, val, sig=15):
+        vals = str(val)
+        if '.' not in vals:
+            vals = vals + '.0000000000000000000'
+        vals = vals.split('.')
+        ints = vals[0]
+        decimals = vals[1][:sig + 1]
+        decimals_rounder = decimals[:-1]
+        decimals = decimals[:sig]
+        if int(decimals_rounder) >= 5:
+            decimals = str(int(decimals) + 1)
+        elif int(decimals_rounder) > 0:
+            decimals = str(int(decimals) - 1)
+        result = ints + '.' + decimals
+        result = result.replace('0E-18', '0')
+        return result
+
+    def validate_type(self, val) -> int or float or Decimal or Fraction:
         """Validates input data types using pydantic type setting. See utils/schema.py for details
 
         Args:
-            val (int or float or decimal): The numerical value to be type checked
+            val (int or float or Decimal or decimal): The numerical value to be type checked
 
         Returns:
-            val (int or float or decimal): the validated type-checked value
+            val (int or float or Decimal or decimal): the validated type-checked value
         """
-        if self.is_solidity:
-            val = Fxp(val, signed=True, n_int=256, n_frac=18)
-        else:
-            data_validator = FloatingPointGeneral
-            input_params = dict(
-                val=val
-            )
-            validated_params = data_validator(**input_params)
-            val = validated_params.val
+        # if self.is_solidity:
+        #     # val = Fxp(val, signed=True, n_int=256, n_frac=18)
+        #     val = Decimal(val)
+        # else:
+        #     # data_validator = FloatingPointGeneral
+        #     # input_params = dict(
+        #     #     val=val
+        #     # )
+        #     # validated_params = data_validator(**input_params)
+        #     # val = validated_params.val
+        # Decimal(val)
         return val
 
+
     def mint_bntkn(self, tkn, x=None, bnt=None):
         """Adds TKN to the available liquidity ledger.
 
         Args:
             tkn (str): The tickername of the token.
-            x (int or float or decimal): The value of liquidity to add. (default=None)
-            bnt (int or float or decimal): The value of bnt liquidity to add. (default=None)
+            x (int or float or Decimal or decimal): The value of liquidity to add. (default=None)
+            bnt (int or float or Decimal or decimal): The value of bnt liquidity to add. (default=None)
 
         Returns:
             y (int): The number of pool tokens issued to the user.
         """
 
-        pool_token_value = self.validate_type(self.get_pool_token_value(tkn))
+        pool_token_value = self.get_pool_token_value(tkn)
+
         if bnt is None:
             x = self.validate_type(x)
             y = self.tokenomics.handle_multiplication(x, pool_token_value)
@@ -485,7 +506,8 @@ class EnvBase(ABC):
         for contract in self.pool_token_supply_ledger:
             self.pool_token_supply_ledger[contract]['block_num'].append(self.block_num)
             if contract == f'bn{tkn}_ERC20_contract':
-                new_val = self.tokenomics.handle_addition(self.pool_token_supply_ledger[contract]['supply'][-1], y)
+                TKN_token_supply = self.validate_type(self.pool_token_supply_ledger[contract]['supply'][-1])
+                new_val = self.tokenomics.handle_addition(TKN_token_supply, y)
                 self.pool_token_supply_ledger[contract]['supply'].append(new_val)
             else:
                 self.pool_token_supply_ledger[contract]['supply'].append(self.pool_token_supply_ledger[contract]['supply'][-1])
@@ -498,7 +520,7 @@ class EnvBase(ABC):
             self
         """
         for TKN in self.whitelisted_tokens:
-            BNTperTKN = self.get_bnt_rate(TKN)
+
             try:
                 if self.vault_ledger[TKN][-1] < 0 or self.available_liquidity_ledger[TKN][-1] < 0 or self.staking_ledger[TKN][-1] < 0 or self.external_protection_wallet_ledger[TKN][-1] < 0:
                     if not warning_only:
@@ -506,8 +528,8 @@ class EnvBase(ABC):
             except:
                 pass
 
+            BNTperTKN = self.get_external_spot_rate(TKN)
             vault_tkn = self.validate_type(self.vault_ledger[TKN][-1])
-            BNTperTKN = self.validate_type(BNTperTKN)
             vault_bnt = self.tokenomics.handle_multiplication(vault_tkn, BNTperTKN)
             if vault_bnt < self.min_liquidity_threshold and (TKN in self.dao_msig_initialized_pools):
                 if self.verbose:
@@ -529,8 +551,10 @@ class EnvBase(ABC):
         bootstrapped_tokens = self.bootstrapped_tokens
         for tkn in self.whitelisted_tokens:
             if tkn != 'BNT':
-                tkn_vault = self.validate_type(self.vault_ledger[tkn][-1])
-                exchange_rate = self.validate_type(self.exchange_rates[tkn])
+                tkn_vault = self.vault_ledger[tkn][-1]
+                exchange_rate = self.exchange_rates[tkn]
+                print('checktype', exchange_rate, tkn_vault)
+
                 tkn_val = self.tokenomics.handle_multiplication(tkn_vault, exchange_rate)
                 if (tkn_val >= self.min_liquidity_threshold) & (tkn not in bootstrapped_tokens) & (tkn in self.dao_msig_initialized_pools):
                     bootstrapped_tokens.append(tkn)
@@ -551,11 +575,11 @@ class EnvBase(ABC):
             satisfies_hlim (bool): Whether the system passes the hlim test.
             satisfies_hmax (bool): Whether the system passes the hmax test.
             reduce_trading_liquidity (bool): Whether the system needs to reduce the available trading liquidity.
-            a (int or float or decimal): the BNT balance of the trading liquidity, as judged from the spot rate.
-            b (int or float or decimal): the TKN balance of the trading liquidity, as judged from the spot rate.
-            c (int or float or decimal): the difference between the TKN balance of the vault, and the TKN trading liquidity.
-            e (int or float or decimal): the TKN balance of the staking ledger.
-            w (int or float or decimal): the TKN balance of the external protection wallet.
+            a (int or float or Decimal or decimal): the BNT balance of the trading liquidity, as judged from the spot rate.
+            b (int or float or Decimal or decimal): the TKN balance of the trading liquidity, as judged from the spot rate.
+            c (int or float or Decimal or decimal): the difference between the TKN balance of the vault, and the TKN trading liquidity.
+            e (int or float or Decimal or decimal): the TKN balance of the staking ledger.
+            w (int or float or Decimal or decimal): the TKN balance of the external protection wallet.
 
         Returns:
             case (str): the human readable logical-case name
@@ -592,17 +616,17 @@ class EnvBase(ABC):
 
         Args:
             tkn (str): The tickername of the token.
-            x (int or float or decimal): The value of liquidity to add. (default=None)
+            x (int or float or Decimal or decimal): The value of liquidity to add. (default=None)
 
         Returns:
-            a (int or float or decimal): the BNT balance of the trading liquidity, as judged from the spot rate.
-            b (int or float or decimal): the TKN balance of the trading liquidity, as judged from the spot rate.
-            c (int or float or decimal): the difference between the TKN balance of the vault, and the TKN trading liquidity.
-            e (int or float or decimal): the TKN balance of the staking ledger.
-            w (int or float or decimal): the TKN balance of the external protection wallet.
+            a (int or float or Decimal or decimal): the BNT balance of the trading liquidity, as judged from the spot rate.
+            b (int or float or Decimal or decimal): the TKN balance of the trading liquidity, as judged from the spot rate.
+            c (int or float or Decimal or decimal): the difference between the TKN balance of the vault, and the TKN trading liquidity.
+            e (int or float or Decimal or decimal): the TKN balance of the staking ledger.
+            w (int or float or Decimal or decimal): the TKN balance of the external protection wallet.
             m (float or decimal): the global system exit fee
             n (float or decimal): the per pool trading fee
-            x (int or float or decimal): the TKN or BNT amount involved in the current action
+            x (int or float or Decimal or decimal): the TKN or BNT amount involved in the current action
         """
         a = self.available_liquidity_ledger[tkn]['BNT'][-1]
         b = self.available_liquidity_ledger[tkn][tkn][-1]
@@ -635,8 +659,8 @@ class EnvBase(ABC):
 
         Args:
             tkn (str): The tickername of the token.
-            bnt_trading_liquidity (int or float or decimal): The value of liquidity to add to bnt.
-            tkn_trading_liquidity (int or float or decimal): The value of liquidity to add to tkn.
+            bnt_trading_liquidity (int or float or Decimal or decimal): The value of liquidity to add to bnt.
+            tkn_trading_liquidity (int or float or Decimal or decimal): The value of liquidity to add to tkn.
 
         Returns:
             self
@@ -661,12 +685,12 @@ class EnvBase(ABC):
 
         Args:
             tkn (str): The tickername of the token.
-            x (int or float or decimal): The value of liquidity to add. (default=None)
-            a (int or float or decimal): the BNT balance of the trading liquidity, as judged from the spot rate.
-            b (int or float or decimal): the TKN balance of the trading liquidity, as judged from the spot rate.
-            c (int or float or decimal): the difference between the TKN balance of the vault, and the TKN trading liquidity.
-            e (int or float or decimal): the TKN balance of the staking ledger.
-            w (int or float or decimal): the TKN balance of the external protection wallet.
+            x (int or float or Decimal or decimal): The value of liquidity to add. (default=None)
+            a (int or float or Decimal or decimal): the BNT balance of the trading liquidity, as judged from the spot rate.
+            b (int or float or Decimal or decimal): the TKN balance of the trading liquidity, as judged from the spot rate.
+            c (int or float or Decimal or decimal): the difference between the TKN balance of the vault, and the TKN trading liquidity.
+            e (int or float or Decimal or decimal): the TKN balance of the staking ledger.
+            w (int or float or Decimal or decimal): the TKN balance of the external protection wallet.
 
         Returns:
             self
@@ -711,7 +735,7 @@ class EnvBase(ABC):
                 self.available_liquidity_ledger[tkn][tkn][-1] = tkntkn
 
             # handle deficit signs
-            elif (self.tokenomics.deficit) & ('ep2' in case):
+            elif (self.tokenomics.deficit) & ('external' in case):
                 tknbnt = self.tokenomics.handle_subtraction(tkn_bnt_liquidity, p)
                 tkntkn = self.tokenomics.handle_subtraction(tkn_tkn_liquidity, r)
                 self.available_liquidity_ledger[tkn]['BNT'][-1] = tknbnt
@@ -867,35 +891,42 @@ class EnvBase(ABC):
 
     def update_exchange_rates(self):
         exchange_rates = self.exchange_rates
-        bnt_price = self.spot_prices['BNT']['price_usd'][-1]
-        for tkn in self.spot_prices:
-            tkn_price = self.spot_prices[tkn]['price_usd'][-1]
-            exchange_rates[tkn] = self.tokenomics.handle_division(tkn_price, bnt_price)
+        bnt_price = self.spot_rates['BNT']['rate'][-1]
+        for tkn in self.spot_rates:
+            tkn_price = self.spot_rates[tkn]['rate'][-1]
+
+            if self.is_solidity:
+                exchange_rates[tkn] = Fraction(tkn_price, bnt_price)
+            else:
+                exchange_rates[tkn] = tkn_price / bnt_price
         self.exchange_rates = exchange_rates
 
-    def get_pool_token_value(self, tkn):
+    def get_pool_token_value(self, tkn) -> int or float or Decimal:
         """Get pool token value in bnTKN/TKN. If this is the first issuance, the pool token value is forced to 1.
 
         Args:
-            x (int or float or decimal): The value of BNT to add to the vortex
+            x (int or float or Decimal or decimal): The value of BNT to add to the vortex
         Returns:
-            bnTKN_value (int or float or decimal): value of a single bnTKN pool token.
+            bnTKN_value (int or float or Decimal or decimal): value of a single bnTKN pool token.
         """
         staked_amount = self.staking_ledger[tkn][-1]
         bnTKN_supply = self.pool_token_supply_ledger[f'bn{tkn}_ERC20_contract']['supply'][-1]
         if staked_amount > 0 and bnTKN_supply > 0:
-            bnTKN_value = self.tokenomics.handle_division(bnTKN_supply, staked_amount)
+            if self.is_solidity:
+                bnTKN_value = Fraction(bnTKN_supply, staked_amount)
+            else:
+                bnTKN_value = bnTKN_supply / staked_amount
         else:
             bnTKN_value = 1
         return bnTKN_value
 
-    def get_spot_rate(self, tkn):
+    def get_internal_spot_rate(self, tkn) -> int or float or Decimal:
         """Gets the spot rate in units of BNT/TKN as determined by the trading liquidity balances of the pool
 
         Args:
             tkn (str): The value of BNT to add to the vortex
         Returns:
-            spot_rate (int or float or decimal): spot rate in units of BNT/TKN
+            spot_rate (int or float or Decimal or decimal): spot rate in units of BNT/TKN
         """
         if tkn == 'BNT':
             spot_rate = self.exchange_rates[tkn]
@@ -903,28 +934,34 @@ class EnvBase(ABC):
             TKN_amt = self.available_liquidity_ledger[tkn][tkn][-1]
             BNT_amt = self.available_liquidity_ledger[tkn]['BNT'][-1]
             if TKN_amt > 0 and BNT_amt > 0:
-                spot_rate = self.tokenomics.handle_division(BNT_amt, TKN_amt)
+                if self.is_solidity:
+                    spot_rate = Fraction(BNT_amt, TKN_amt)
+                else:
+                    spot_rate = BNT_amt / TKN_amt
             else:
                 spot_rate = self.exchange_rates[tkn]
 
         return spot_rate
 
-    def get_bnt_rate(self, tkn) -> float:
+    def get_external_spot_rate(self, tkn) -> int or float or Decimal or Fraction:
         """Gets the rate in units of TKN/BNT as determined by the external price feed
 
         Args:
             tkn (str): The token tickername.
         Returns:
-            bnt_rate (int or float or decimal): bnt rate in units of TKN/BNT
+            bnt_rate (int or float or Decimal or decimal): bnt rate in units of TKN/BNT
         """
-        bnt_price = self.external_price_feed['BNT']['price_usd'][-1]
-        tkn_price = self.external_price_feed[tkn]['price_usd'][-1]
-        bnt_price = self.validate_type(bnt_price)
-        tkn_price = self.validate_type(tkn_price)
-        bnt_rate = self.tokenomics.handle_division(tkn_price, bnt_price)
+        bnt_price = self.external_rate_feed['BNT']['rate'][-1]
+        tkn_price = self.external_rate_feed[tkn]['rate'][-1]
+        # bnt_price = self.validate_type(bnt_price)
+        # tkn_price = self.validate_type(tkn_price)
+        if self.is_solidity:
+            bnt_rate = Fraction(tkn_price, bnt_price)
+        else:
+            bnt_rate = tkn_price / bnt_price
         return bnt_rate
 
-    def dao_msig_initialize_pools(self, force=False):
+    def dao_msig_initialize_pools(self, force=False, multiplier=1):
         """Command to initialize the pools in simulation of DAOmsig actions.
 
         Returns:
@@ -932,8 +969,10 @@ class EnvBase(ABC):
         """
         dao_msig_initialized_pools = self.dao_msig_initialized_pools
         for TKN in self.whitelisted_tokens:
-            BNTperTKN = self.get_bnt_rate(TKN)
-            BNTperTKN = self.validate_type(BNTperTKN)
+
+            BNTperTKN = self.get_external_spot_rate(TKN)
+
+            # BNTperTKN = self.validates_type(BNTperTKN)
 
             tkn_tkn_liquidity = self.available_liquidity_ledger[TKN][TKN][-1]
             tkn_tkn_liquidity = self.validate_type(tkn_tkn_liquidity)
@@ -960,12 +999,13 @@ class EnvBase(ABC):
 
             zero = self.validate_type(0)
             min_liquidity_threshold = self.validate_type(self.min_liquidity_threshold)
-
+            min_liquidity_threshold = self.tokenomics.handle_multiplication(min_liquidity_threshold, multiplier)
+            print('BNTperTKN ', BNTperTKN, min_liquidity_threshold)
             if (tkn_tkn_liquidity == zero and tkn_bnt_liquidity == zero and tkn_vault_times_bntpertkn > min_liquidity_threshold and (TKN not in self.dao_msig_initialized_pools)) or force:
-
+                # print('checkliq',min_liquidity_threshold, BNTperTKN)
                 self.available_liquidity_ledger[TKN]['block_num'].append(self.available_liquidity_ledger[TKN]['block_num'][-1])
                 self.available_liquidity_ledger[TKN][TKN].append(self.tokenomics.handle_division(min_liquidity_threshold, BNTperTKN))
-                self.available_liquidity_ledger[TKN]['BNT'].append(self.min_liquidity_threshold)
+                self.available_liquidity_ledger[TKN]['BNT'].append(min_liquidity_threshold)
 
                 self.vault_ledger['BNT'].append(self.tokenomics.handle_addition(bnt_vault, min_liquidity_threshold))
                 self.staking_ledger['BNT'].append(self.tokenomics.handle_addition(bnt_staking, min_liquidity_threshold))
@@ -986,9 +1026,9 @@ class EnvBase(ABC):
         Returns:
             self
         """
-        alpha = self.validate_type(self.alpha)
-        spot_rate = self.validate_type(self.get_spot_rate(tkn))
-        current_ema = self.validate_type(self.ema[tkn]['ema'][-1])
+        alpha = self.alpha
+        spot_rate = self.get_internal_spot_rate(tkn)
+        current_ema = self.ema[tkn]['ema'][-1]
         ema = self.ema
         if tkn in self.ema:
             if ema[tkn]['block_num'][-1] < self.block_num:
@@ -1000,41 +1040,53 @@ class EnvBase(ABC):
             ema[tkn]['ema'].append(self.exchange_rates[tkn])
         self.ema = ema
 
-    def is_within_ema_tolerance(self, tkn, lower_limit=.99, upper_limit=1.01):
+    def is_within_ema_tolerance(self, tkn, lower_limit=Fraction(99, 100), upper_limit=Fraction(101, 100)):
         """Check if the spot rate is outside of the EMA tolerance. No changes are made to the pool depth if outside limits.
 
         Args:
             tkn (str): The tickername of the token.
         Returns:
-            bnTKN_value (int or float or decimal): value of a single bnTKN pool token.
+            bnTKN_value (int or float or Decimal or decimal): value of a single bnTKN pool token.
         """
-        spot_rate = self.get_spot_rate(tkn)
-        spot_rate = self.validate_type(spot_rate)
-
+        spot_rate = self.get_internal_spot_rate(tkn)
         ema = self.ema[tkn]['ema'][-1]
-        ema = self.validate_type(ema)
-
-        lower_limit = self.validate_type(lower_limit)
-        upper_limit = self.validate_type(upper_limit)
-
+        # print('emacheck',ema, spot_rate)
         lower_bound = self.tokenomics.handle_multiplication(lower_limit, ema)
         upper_bound = self.tokenomics.handle_multiplication(upper_limit, ema)
 
         result = lower_bound <= spot_rate <= upper_bound
         if self.verbose:
             print(
+                f"block_num={self.block_num} \n",
                 f"spot_rate={spot_rate} \n",
                 f"ema={ema} \n",
-                f"EMA Test = {result}, ({lower_limit} * {ema}) <= {spot_rate} <= ({upper_limit} * {ema}), tkn={tkn}"
+                f"EMA Test = {result}, ({lower_limit} * {ema}) <= {spot_rate} <= ({upper_limit} * {ema}), tkn={tkn} \n",
             )
         return result
 
+    def remove_tkn_from_vault(self, tkn, x):
+        """Adds x tkn to the self.vault_ledger dictionary with the same block_num number.
+
+        Args:
+            tkn (str): The tickername of the token.
+            x (int or float or Decimal or decimal): The amount of tkn.
+        Returns:
+            self
+        """
+        for key in self.vault_ledger:
+            if key == tkn:
+                new_val = self.tokenomics.handle_subtraction(self.vault_ledger[key][-1], x)
+                self.vault_ledger[key].append(new_val)
+            else:
+                self.vault_ledger[key].append(self.vault_ledger[key][-1])
+            self.vault_ledger['block_num'][-1] = self.block_num
+
     def add_tkn_to_vault(self, tkn, x):
         """Adds x tkn to the self.vault_ledger dictionary with the same block_num number.
 
         Args:
             tkn (str): The tickername of the token.
-            x (int or float or decimal): The amount of tkn.
+            x (int or float or Decimal or decimal): The amount of tkn.
         Returns:
             self
         """
@@ -1051,7 +1103,7 @@ class EnvBase(ABC):
 
         Args:
             tkn (str): The tickername of the token.
-            x (int or float or decimal): The amount of tkn.
+            x (int or float or Decimal or decimal): The amount of tkn.
         Returns:
             self
         """
@@ -1063,11 +1115,11 @@ class EnvBase(ABC):
                 self.staking_ledger[key].append(self.staking_ledger[key][-1])
             self.staking_ledger['block_num'][-1] = self.block_num
 
-    def add_tkn_to_vortex_ledger(self, x):
+    def add_tkn_to_vortex_ledger(self, tkn, x):
         """Adds TKN to the vortex ledger.
 
         Args:
-            x (int or float or decimal): The value of BNT to add to the vortex.
+            x (int or float or Decimal or decimal): The value of BNT to add to the vortex.
 
         Returns:
             self
@@ -1075,6 +1127,7 @@ class EnvBase(ABC):
         new_val = self.tokenomics.handle_addition(self.vortex_ledger['BNT'][-1], x)
         self.vortex_ledger['BNT'].append(new_val)
         self.vortex_ledger['block_num'].append(self.block_num)
+        self.available_liquidity_ledger[tkn]['funding_limit'][-1] += new_val
 
     def add_tkn_to_pool_token_supply_ledger(self, tkn, x):
         self.pool_token_supply_ledger[tkn]['block_num'].append(self.block_num)
@@ -1090,7 +1143,7 @@ class EnvBase(ABC):
 
         Args:
             tkn (str): The tickername of the token.
-            x (int or float or decimal): The value of liquidity to add.
+            x (int or float or Decimal or decimal): The value of liquidity to add.
 
         Returns:
             self
@@ -1098,11 +1151,11 @@ class EnvBase(ABC):
         for key in self.available_liquidity_ledger:
             if key == tkn:
 
-                a = self.available_liquidity_ledger[key]['BNT'][-1]
-                b = self.available_liquidity_ledger[key][tkn][-1]
-                c = self.tokenomics.handle_subtraction(self.vault_ledger[key][-1], b)
-                spot_rate = self.get_spot_rate(tkn)
-                funding_limit = self.available_liquidity_ledger[key]['funding_limit'][-1]
+                a = self.validate_type(self.available_liquidity_ledger[key]['BNT'][-1])
+                b = self.validate_type(self.available_liquidity_ledger[key][tkn][-1])
+                c = self.validate_type(self.tokenomics.handle_subtraction(self.vault_ledger[key][-1], b))
+                spot_rate = self.get_internal_spot_rate(tkn)
+                funding_limit = self.validate_type(self.available_liquidity_ledger[key]['funding_limit'][-1])
 
                 print(f'a={a}, b={b}, c={c}, spot_rate={spot_rate}, funding_limit={funding_limit}')
 
@@ -1151,8 +1204,6 @@ class EnvBase(ABC):
                 self.available_liquidity_ledger[key]['block_num'].append(self.block_num)
                 final_liquidity = self.available_liquidity_ledger[key]['BNT'][-1]
                 delta = self.tokenomics.handle_subtraction(final_liquidity, a)
-                self.add_tkn_to_vault('BNT', delta)
-                self.add_tkn_to_staking_ledger('BNT', delta)
 
                 bnt = self.staking_ledger['BNT'][-1]
                 bnt_supply = self.pool_token_supply_ledger[f"bnBNT_ERC20_contract"]["supply"][-1]
@@ -1161,6 +1212,9 @@ class EnvBase(ABC):
                     update_amt = self.tokenomics.handle_division(delta, bnt_bnbnt_rate)
                     self.add_tkn_to_pool_token_supply_ledger("bnBNT_ERC20_contract", update_amt)
 
+                self.add_tkn_to_vault('BNT', delta)
+                self.add_tkn_to_staking_ledger('BNT', delta)
+
                 if "bnBNT" not in self.protocol_agent.wallet:
                     self.protocol_agent.wallet["bnBNT"] = {'block_num': [0], "bnBNT": [0]}
 
