@@ -19,6 +19,12 @@ class AvailableLiquidityLedger(LedgerBase):
                 'BNT': [0],
                 'ETH': [0]
             },
+            'DAI': {
+                'block_num': [0],
+                'funding_limit': [0],
+                'BNT': [0],
+                'DAI': [0]
+            },
             'wBTC': {
                 'block_num': [0],
                 'funding_limit': [0],
